from ._optimizer import approximate_local_min, approximate_local_min_axenie
from .learning_rate_functions import *
from .plotting import *
from .spice_net import SpiceNet
from .spice_net_hcm import SpiceNetHcm
from .spice_net_som import SpiceNetSom



import math

import numpy as np

from .learning_rate_functions import LearningRateFunction


class SpiceNetSom:
    """
    This self organizing map implementation is specifically for SpiceNet.
    For more information read the theory papers included.
    """

    def __init__(self,
                 n_neurons: int,
                 value_range_start: float,
                 value_range_end: float,
                 lrf_tuning_curve: LearningRateFunction,
                 lrf_interaction_kernel: LearningRateFunction,
                 key: str,
                 dimensions: int = 1, ):
        """
        Creates a SpiceNetSom with neurons equally distributed across the specified value range.

        :param n_neurons: The number of neurons that should be created.
        :param value_range_start: Start of the believed value range. (This is only relevant for the initialisation.
        It is possible to fit values outside of this range!)
        :param value_range_end: End of the believed value range. (This is only relevant for the initialisation.
        It is possible to fit values outside of this range!)
        :param lrf_tuning_curve: Use this parameter to define how fast the tuning curve is changed.
        :param lrf_interaction_kernel: Use this parameter to define how strong the interaction kernel affects the preferred value of a node.
        :param key: The key to identify the som.
        :param dimensions: The dimensions of the som nodes (the data represented).
        """
        assert dimensions >= 1

        self.__key = key
        self.__lrf_interaction_kernel = lrf_interaction_kernel
        self.__lrf_tuning_curve = lrf_tuning_curve
        self.__iteration = 0

        distance = value_range_end - value_range_start
        step_size = distance / n_neurons

        self.__neurons = []
        pos = value_range_start + step_size / 2.0

        for i in range(n_neurons):
            if dimensions == 1:
                new_neuron = SpiceNetSom.__SomNeuron(pos)
            else:
                new_neuron = SpiceNetSom.__SomNeuron(np.array([pos for i in range(dimensions)]))
            self.__neurons.append(new_neuron)
            pos += step_size

    def __len__(self) -> int:
        return len(self.__neurons)

    def print_neurons(self):
        for neuron in self.__neurons:
            print(neuron)

    def fit(self, values: list[float | np.ndarray[any, np.dtype[np.float64]]], epochs: int):
        for epoch in range(epochs):
            for i in range(len(values)):
                winning_neuron_index, _ = self.__argmax_neuron_activation(values[i])

                for j in range(len(self.__neurons)):
                    # TODO: Fix passed iterations
                    self.__neurons[j].update(values[i],
                                             self.__lrf_tuning_curve.call(self.__iteration),
                                             self.__lrf_interaction_kernel.call(self.__iteration),
                                             j - winning_neuron_index)
                self.__iteration += 1

    def get_key(self):
        return self.__key

    def get_as_matrix(self) -> np.ndarray:
        """
        Returns all neurons like a table with column 0 representing the preferred value
        and column 1 representing the tuning curve width.
        :return: A numpy matrix.
        """
        return np.array([[neuron.preferred_value for neuron in self.__neurons],
                         [neuron.tuning_curve_width for neuron in self.__neurons]]).transpose()

    def get_activation_vector(self, value: float | np.ndarray[any, np.dtype[np.float64]]) -> np.array:
        return np.array([neuron.activation_for_value(value) for neuron in self.__neurons])

    def calculate_activation_values(self, values: list[float | np.ndarray[any, np.dtype[np.float64]]]):
        """
        Calculates activation values for all neurons in the som.
        :param values:
        :return: A numpy array the first col is the preferred value, second col is the tuning curve width
        and the following cols are the activation values.
        """
        activation_values = np.array([neuron.activation_for_values(values) for neuron in self.__neurons])
        return np.concatenate((
            np.array(
                [[neuron.preferred_value for neuron in self.__neurons],
                 [neuron.tuning_curve_width for neuron in self.__neurons]]).transpose(),
            activation_values),
            axis=1)

    def get_winning_neuron_index(self, value: float | np.ndarray[any, np.dtype[np.float64]]) -> (int, float):
        winning_neuron_index, activation_values = self.__argmax_neuron_activation(value)
        return winning_neuron_index, activation_values[winning_neuron_index]

    def naive_decode(self, value: float, neuron_index: int) -> float:
        # activation_value = self.__neurons[neuron_index].activation_for_value(value)
        activation_value = value
        # activation_value += 1000000
        # print(f'neuron: {neuron_index} value: {value} activation: {activation_value}')
        # print(f'prefered_value: {self.__neurons[neuron_index].preferred_value}')
        # print(f'width: {self.__neurons[neuron_index].tuning_curve_width}')
        # print(self.__neurons[neuron_index].tuning_curve_width)
        # print(math.sqrt(2 * math.pi) * activation_value * self.__neurons[neuron_index].tuning_curve_width ** 2)
        # print(2 * self.__neurons[neuron_index].tuning_curve_width ** 2 * math.log(
        #     math.sqrt(2 * math.pi) * activation_value * self.__neurons[neuron_index].tuning_curve_width ** 2))
        r = math.sqrt(2 * self.__neurons[neuron_index].tuning_curve_width ** 2 * math.log(
            math.sqrt(2 * math.pi) * activation_value * self.__neurons[neuron_index].tuning_curve_width ** 2, 10))

        if neuron_index < len(self.__neurons) / 2:
            return self.__neurons[neuron_index].preferred_value - r
        else:
            return self.__neurons[neuron_index].preferred_value + r

    def __argmax_neuron_activation(self, value: float | np.ndarray[any, np.dtype[np.float64]]):
        """
        Calculates the neuron with the highest activation value.
        :param value: The value for wich the activation values should be calculated.
        :return: The index of the winning neuron
        and a dictionary containing the index of a neuron with the calculated activation value.
        """
        winning_neuron_index: int = 0
        max_activation = self.__neurons[0].activation_for_value(value)
        activation_dict = {0: max_activation}

        for i in range(1, len(self.__neurons)):
            new_activation = self.__neurons[i].activation_for_value(value)
            # add new value to a dictionary so the activations only have to be calculated a single time
            activation_dict[i] = new_activation

            if new_activation > max_activation:
                max_activation = new_activation
                winning_neuron_index = i
        return winning_neuron_index, activation_dict

    class __SomNeuron:
        """
        This object represents a som neuron. Since this som is 1D the neuron only knows th previous and next neuron.
        """

        def __init__(self,
                     preferred_value: float | np.ndarray[any, np.dtype[np.float64]],
                     tuning_curve_width: float = 0.001):
            """
            Creates a som neuron.
            :param preferred_value: The "expected value" of the Normal distribution
            representing the activation function.
            :param tuning_curve_width: The width of the tuning curve (the Normal distribution).
            """
            self.preferred_value = preferred_value
            """ The "expected value" of the Normal distribution representing the activation function. """
            self.tuning_curve_width = tuning_curve_width
            """ The width of the tuning curve (the Normal distribution). """

        def __str__(self):
            return f'SomeNeuron: preferred_value={self.preferred_value}, tuning_curve_width={self.tuning_curve_width}'

        def activation_for_value(self, value: float | np.ndarray[any, np.dtype[np.float64]]):
            """
            Returns the activation value of the neuron.
            :param value: The value for wich the activation has to be calculated.
            :return: The "height" of the tuning curve at the position of the value.
            """
            if isinstance(value, float):
                return (
                        (1.0 / (math.sqrt(2.0 * math.pi) * self.tuning_curve_width))
                        *
                        math.exp(
                            (-((value - self.preferred_value) ** 2)) / #FIXME: Wo muss das - hin vor dem exponent oder nach
                            (2.0 * self.tuning_curve_width ** 2))
                )
            else:
                return (
                        (1.0 / (math.sqrt(2.0 * math.pi) * self.tuning_curve_width))
                        *
                        math.exp(
                            (-((np.linalg.vector_norm(value - self.preferred_value)) ** 2)) /
                            (2.0 * self.tuning_curve_width ** 2))
                )

        def activation_for_values(self, values: list[float | np.ndarray[any, np.dtype[np.float64]]]):
            """
            Returns the activation values for a list of values.
            :param values: The values for wich the activation has to be calculated.
            :return: An array of activation values, ordered like the input list.
            """
            return np.array([self.activation_for_value(value) for value in values])

        def update(self, value: float | np.ndarray[any, np.dtype[np.float64]], learn_rate: float,
                   interaction_kernel_learning_rate: float,
                   distance_to_winner: int):
            """
            Updates the weights of the neurons.
            :param value: The value in wich "direction" the neuron should move.
            :param learn_rate: The learn rate. (This value should be depending on the iteration.)
            :param interaction_kernel_learning_rate: The sigma of the interaction kernel.
            (This value should be depending on the iteration.)
            :param distance_to_winner: The distance between this neuron and the winning neuron, in the SOM.
            Here are not the weights / postions in the value range relevant.
            (N1 weight: 31.9) --- (N2 weight: 32) --- (N3 weight: 32.4) --- (N4 weight: 33)
            The Distance of N4 to N1 is 3
            :return:
            """
            interaction_kernel_value = math.exp(
                (-abs(distance_to_winner) ** 2) / (2 * interaction_kernel_learning_rate ** 2))
            old_preferred_value = self.preferred_value #FIXME: Wirklich mit old arbeiten?
            self.preferred_value += learn_rate * interaction_kernel_value * (value - self.preferred_value)
            if isinstance(value, float):
                self.tuning_curve_width += learn_rate * interaction_kernel_value * (
                        (value - old_preferred_value) ** 2 - self.tuning_curve_width ** 2
                )
            else:
                self.tuning_curve_width += learn_rate * interaction_kernel_value * (
                        np.linalg.vector_norm(value - old_preferred_value) ** 2 - self.tuning_curve_width ** 2
                )



from typing import Optional

import numpy as np
import pandas as pd
import plotly.express as px
from matplotlib import pyplot as plt

from ..spice_net_hcm import SpiceNetHcm
from ..spice_net_som import SpiceNetSom


def plot_som(som: SpiceNetSom, select_all_button: bool = False):
    stem_x_values = som.get_as_matrix()[:, 0]
    x_min = stem_x_values.min()
    x_max = stem_x_values.max()
    distance = x_max - x_min
    x_min -= distance * 0.15
    x_max += distance * 0.15
    x_plt_vals = np.linspace(x_min, x_max, 500)
    activation_values = som.calculate_activation_values(x_plt_vals.tolist())
    df = pd.DataFrame(activation_values[:, 2:].transpose(), index=x_plt_vals)
    fig = px.line(df)
    if select_all_button:
        fig.update_layout(dict(updatemenus=[
            dict(
                type="buttons",
                direction="left",
                buttons=list([
                    dict(
                        args=["visible", "legendonly"],
                        label="Deselect All",
                        method="restyle"
                    ),
                    dict(
                        args=["visible", True],
                        label="Select All",
                        method="restyle"
                    )
                ]),
                pad={"r": 10, "t": 10},
                showactive=False,
                x=1,
                xanchor="right",
                y=1.1,
                yanchor="top"
            ),
        ]
        ))
    fig.show(title='Activation values')


def plot_hcm(hcm: SpiceNetHcm,
             normed: bool = False,
             filter_start: Optional[float] = None,
             filter_end: Optional[float] = None,
             width=None,
             height=None):
    if len(hcm.get_soms()) == 2:
        matrix = hcm.get_matrix() if normed is False else np.where(hcm.get_matrix() > 0, 1, -1)
        fig = px.imshow(matrix, text_auto=False, aspect="auto", width=width, height=height)
        fig['layout']['yaxis']['autorange'] = "min"
        fig.update_layout(yaxis=dict(scaleanchor='x'))
        fig.show()
    elif len(hcm.get_soms()) == 3:
        it = np.nditer(hcm.get_matrix(), flags=['multi_index'])
        x_list = []
        y_list = []
        z_list = []
        value_list = []
        for value in it:
            value_list.append(float(value))
            x_list.append(it.multi_index[0])
            z_list.append(it.multi_index[1])
            y_list.append(it.multi_index[2])
        df = pd.DataFrame(data={'x': x_list, 'y': y_list, 'z': z_list, 'value': value_list})
        if filter_start is not None:
            df = df[df['value'] >= filter_start]
        if filter_end is not None:
            df = df[df['value'] <= filter_end]
        fig = px.scatter_3d(df, x='x', y='y', z='z',
                            color='value', opacity=0.7,
                            width=width, height=height)
        fig.update_traces(marker_size=3)
        fig.show()
    else:
        raise ValueError('Can only plot 2 or 3 dimensional correlation matrixes')


def plot_som_weights_stem(som: SpiceNetSom):
    stem_x_values = som.get_as_matrix()[:, 0]
    stem_y_values = som.get_as_matrix()[:, 1]

    plt.stem(stem_x_values, stem_y_values)
    plt.show()


def plot_som_comparison_to_data(som: SpiceNetSom, data: list, n_bins: Optional[int] = None):
    bins = n_bins if n_bins is not None else len(som)
    stem_x_values = som.get_as_matrix()[:, 0]
    stem_y_values = som.get_as_matrix()[:, 1]

    fig, axs = plt.subplots(2, 2, sharey=False, tight_layout=True)
    fig.set_size_inches(18.5, 21.0, forward=True)

    axs[0, 0].stem(stem_x_values, stem_y_values)

    # first row of plots
    xmax = axs[0, 0].viewLim.xmax
    xmin = axs[0, 0].viewLim.xmin

    # second row of plots
    x_plt_vals = np.linspace(xmin, xmax, 500)
    activation_values = som.calculate_activation_values(x_plt_vals)
    num_rows, num_cols = activation_values.shape
    for i in range(num_rows):
        axs[1, 0].plot(x_plt_vals, activation_values[i, 2:], linewidth=2)
    axs[0, 1].hist(data, bins=bins)
    axs[1, 1].hist(data, bins=bins)
    plt.show()



import typing

import numpy as np

from .learning_rate_functions import LearningRateFunction
from .spice_net_som import SpiceNetSom


class SpiceNetHcm:

    def __init__(self,
                 soms: typing.Collection[SpiceNetSom],
                 lrf_weights: LearningRateFunction,
                 lrf_trust_of_new: LearningRateFunction):
        """
        This object represents a hebbian correlation matrix between 2 SOMs.
        :param soms: A collection of SpiceNetSom objects mapped to a string key.
        :param lrf_weights: The learning rate of the weights, how strong each update affects the matrix.
        :param lrf_trust_of_new: This implementation uses the hebbian covariance learning mechanism. This uses an 'average value', this value is update continuous in this implementation. Use this parameter to tell how much you trust that the new value and want it to change the average. For more information on the mechanism look here: https://rkypragada.medium.com/hebbian-learning-c2166ac0f48d
        """
        self.__lrf_trust_of_new = lrf_trust_of_new
        self.__lrf_weights = lrf_weights
        self.__soms: list[SpiceNetSom] = [som for som in soms]

        index = 0
        # self.__activation_bars: list[np.ndarray] = []
        self.__activation_bars: list[np.ndarray] = [np.zeros(len(som)) for som in soms]
        self.__current_activations: list[np.ndarray] = [np.zeros(len(som)) for som in soms]
        self.__index_to_som: dict[int, SpiceNetSom] = {}
        self.__som_key_to_index: dict[str, int] = {}
        for som in soms:
            self.__index_to_som[index] = som
            self.__index_to_som[som.get_key()] = index
            # self.__activation_bars.append(np.zeros(len(som)))
            self.__som_key_to_index[som.get_key()] = index
            index += 1
        self.__weights = np.ones([len(som) for som in soms])
        self.__iteration = 0

    def get_soms(self):
        return self.__soms

    def get_matrix(self) -> np.ndarray:
        """
        Get the weight matrix.
        :return: The matrix y-axis is the first som given to the constructor (x-axis the second one).
        """
        return self.__weights

    def calculate_should_pattern(self, target_som_key: str,
                                 input_activations: typing.Mapping[str, np.ndarray]) -> np.ndarray:
        """
        The anticipated activation pattern.
        :param target_som_key:
        :param input_activations:
        :return: This function returns a relative representation of the activation.
        These are not the numbers return by the action function, but the ratio of the activation values is close to the real value.
        """
        target_index = self.__som_key_to_index[target_som_key]
        sorted_vectors = []
        for som in self.__soms:
            if som.get_key() == target_som_key:
                continue
            if input_activations[som.get_key()] is None and som.get_key() != target_som_key:
                raise ValueError('Missing value for ' + som.get_key())
            else:
                sorted_vectors.append(input_activations[som.get_key()])

        result = []
        for i in range(len(self.__soms[target_index])):
            array_slice = np.take(self.__weights, indices=i, axis=target_index)
            grids = np.meshgrid(*sorted_vectors, indexing='ij')
            for grid in grids:
                array_slice *= grid
            result.append(np.sum(array_slice))
        return np.array(result)

        # TODO: change to relative collapse instead of absolute
        result = []
        for i in range(len(self.__soms[target_index])):
            index_list = [0 for _ in range(len(self.__soms))]
            index_list[target_index] = i
            keys = [key for key in input_activations.keys()]
            result.append(self.__should_help(keys, index_list, [], input_activations))
        return np.array(result)

    def __should_help(self, remaining_keys: list[str],
                      indices: list[int],
                      values: list[float],
                      input_activations: typing.Mapping[str, np.ndarray]) -> float:
        """

        :param remaining_keys:
        :param indices:
        :param values:
        :param input_activations:
        :return:
        """
        if len(remaining_keys) > 0:
            current_key = remaining_keys.pop()
            som_weight_index = self.__som_key_to_index[current_key]
            result = 0.0
            for i in range(len(input_activations[current_key])):
                val = input_activations[current_key][i]
                ind = indices.copy()
                ind[som_weight_index] = i
                result += self.__should_help(remaining_keys, ind, values + [val], input_activations)
            return result
        else:
            result = 1.0
            for value in values:
                result *= value
            return self.__weights[*indices] * result

    def fit(self,
            values: typing.Mapping[str, list[float | np.ndarray[any, np.dtype[np.float64]]]],
            epochs: int):
        """
        Fits the hebbian correlation matrix.

        :param values:
        :param epochs: How many times the matrix is fitted to the data.
        """

        data_len = -1
        for som in self.__soms:
            if values[som.get_key()] is None:
                raise ValueError('Missing value for ' + som.get_key())
            if data_len != -1 and data_len != len(values[som.get_key()]):
                raise ValueError('The length of the data columns is not allways the same.')
            if data_len == -1:
                data_len = len(values[som.get_key()])

        for _ in range(epochs):
            for i in range(data_len):
                l = []
                for som in self.__soms:
                    index_in_weight_matrix = self.__som_key_to_index[som.get_key()]
                    activation_vector = som.get_activation_vector(values[som.get_key()][i])
                    self.__current_activations[index_in_weight_matrix] = activation_vector
                    self.__activation_bars[index_in_weight_matrix] = (
                            (1.0 - self.__lrf_trust_of_new.call(self.__iteration))
                            * self.__activation_bars[index_in_weight_matrix]
                            + self.__lrf_trust_of_new.call(self.__iteration)
                            * activation_vector)
                    l.append(activation_vector - self.__activation_bars[index_in_weight_matrix])
                shape_of_weight = np.shape(self.__weights)
                # with np.nditer(self.__weights, op_flags=['readwrite'], flags=['multi_index']) as it:
                #    for x in it:
                #        x[...] += self.__calculate_delta_value(it.multi_index)
                #        #print("%d <%s>" % (x, it.multi_index))

                delta_matrix = self.__generate_n_dimensional_tensor(l)
                self.__weights += delta_matrix * self.__lrf_weights.call(self.__iteration)
                #self.__iterate_over_matrix(shape_of_weight, self.__update_weight)
                self.__iteration += 1

    def __generate_n_dimensional_tensor(self, vectors):
        """
        Erzeugt einen n-dimensionalen Tensor aus n Vektoren,
        wobei jede Dimension die Multiplikation der Elemente aus den Vektoren darstellt.

        :param vectors: Liste von 1D-Arrays (Vektoren)
        :return: n-dimensionaler Tensor
        """
        # Konvertiere alle Vektoren in NumPy-Arrays
        vectors = [np.array(vec) for vec in vectors]

        # Erstelle die Ausgangsform für Broadcasting
        shape = [len(vec) for vec in vectors]
        grids = np.meshgrid(*vectors, indexing='ij')

        # Multipliziere alle Gitterpunkt-Werte
        result = np.ones(shape)
        for grid in grids:
            result *= grid  # Elementweise Multiplikation

        return result

    def __update_weight(self, ind: list[int]):
        # print(ind)
        # print(self.__weights[*ind])
        self.__weights[*ind] += self.__calculate_delta_value(ind)

    def __calculate_delta_value(self, ind: list[int]) -> float:
        result = 1.0
        # print(f'weightpos: {ind}')
        for i in range(len(ind)):
            # print([i, ind[i]])
            # print(self.__current_activations[i][ind[i]])
            # print(self.__activation_bars[i][ind[i]])
            result *= (self.__current_activations[i][ind[i]] - self.__activation_bars[i][ind[i]])
            # result *= self.__activation_bars[ind[i]]
        result *= self.__lrf_weights.call(self.__iteration)
        # print(f'index: {ind} delta value: {result}')
        return result

    def __iterate_over_matrix_iterative(self, shape, call_back_fn):
        stack = [[]]  # Start mit einer leeren Liste für die Indizes

        while stack:
            indices = stack.pop()

            # Wenn die Länge der Indizes der Dimension der Matrix entspricht, ist das Ziel erreicht
            if len(indices) == len(shape):
                call_back_fn(indices)
            else:
                # Sonst füge weitere Indizes hinzu
                current_dimension = len(indices)
                for i in range(shape[current_dimension]):
                    stack.append(indices + [i])

    def __iterate_over_matrix(self, shape: typing.Tuple[int], call_back_fn: typing.Callable[[list[int]], any],
                              indices=None, current_index=0):
        if indices is None:
            indices = []
        ind = indices.copy()
        if current_index == len(shape):
            call_back_fn(indices)
            return
        new_current_index = current_index + 1
        for i in range(shape[current_index]):
            self.__iterate_over_matrix(shape, call_back_fn, indices=ind + [i], current_index=new_current_index)



from .learning_rate_function import LearningRateFunction


class ConstLRF(LearningRateFunction):
    def __init__(self, value: float):
        self.__value = value

    def call(self, iteration: int) -> float:
        return self.__value



import math

from .learning_rate_function import LearningRateFunction


class ExpEDecayLRF(LearningRateFunction):
    def __init__(self, speed: float, approached_value: float, x_shift: float = 0.0):
        """
        (e^(-iteration * speed)) + bias
        :param speed:
        :param approached_value:
        """
        self.__bias: float = approached_value
        self.__speed: float = speed
        self.__x_shift: float = x_shift

    def call(self, iteration: int) -> float:
        return math.exp(-iteration * self.__speed + self.__x_shift) + self.__bias



from .learning_rate_function import LearningRateFunction


class LinearLRF(LearningRateFunction):

    def __init__(self, slope: float, bias: float):
        self.__slope = slope
        self.__bias = bias

    def call(self, iteration: int) -> float:
        return self.__slope * iteration + self.__bias


def declining_linear_lrf_from_interval(start: float, stop: float, steps: int) -> LinearLRF:
    bias = start
    slope = -((start - stop) / steps)
    return LinearLRF(slope, bias)



from . import LearningRateFunction


class InverseTimeAdaptation(LearningRateFunction):

    def __init__(self,
                 init_value: float,
                 final_value: float,
                 planned_iterations: int,
                 iteration_0=1):
        self.__planned_iterations = planned_iterations
        self.__B = ((final_value * planned_iterations - init_value * iteration_0)
                    / init_value - final_value)
        self.__A = init_value * iteration_0 + self.__B * iteration_0

    def call(self, iteration: int) -> float:
        t = self.__planned_iterations - iteration
        return self.__A / (t + self.__B)



from typing import Callable

__c = 0.381966  # c = (3 - sqrt(5)) / 2
__eps = 1.1e-16


def approximate_local_min(interval_start: float,
                          interval_end: float,
                          function: Callable[[float], float],
                          tolerance: float = None,
                          eps: float = None,
                          t: float = None) -> float:
    """
    This function approximates the local minima of the given function in an interval. For more details look up ISBN 978-1-306-35261-1.

    :param tolerance: The tolerance for the approximation.
    :param interval_start: The included beginning of the interval.
    :param interval_end: the included end of the interval.
    :param eps: t and eps define a tolerance tol − eps | x | + t, and f is never evaluated at two points closer together than tol.
    :param t: t and eps define a tolerance tol − eps | x | + t, and f is never evaluated at two points closer together than tol.
    :param function: A function defined on the interval. The local minima will be approximated for this function.
    """
    if not (tolerance is not None or (eps is not None and t is not None)):
        raise TypeError("Either tolerance or (eps and t) must be defined.")

    print(f'start: {interval_start} end: {interval_end}')
    v = w = x = interval_start + __c * (interval_end - interval_start)
    e = 0
    fv = fw = fx = function(x)
    # print(fv)

    # tol = 1.0e-6 * (limL + limH) / 2.0
    tol_radius = tolerance if tolerance is not None else eps * abs(x) + t
    tol_both_directions = 2 * tol_radius

    for i in range(0, 50):
        m = 0.5 * (interval_start + interval_end)
        if tolerance is None:
            tol_radius = eps * abs(x)
            tol_both_directions = 2 * tol_radius
        if abs(x - m) > tol_both_directions - 0.5 * (interval_end - interval_start):
            p = q = r = 0
            if abs(e) > tol_radius:
                r = (x - w) * (fx - fv)
                q = (x - v) * (fx - fw)
                p = (x - v) * q - (x - w) * r
                q = 2 * (q - r)
                if q > 0:
                    p = -p
                else:
                    q = -q
                r = e
                e = d
            if abs(p) < abs(0.5 * q * r) and p < q * (interval_start - x) and p < q * (interval_end - x):
                d = p / q
                u = x + d
                if u - interval_start < tol_both_directions or interval_end - u < tol_both_directions:
                    d = tol_radius if x < m else -tol_radius
            else:
                e = (interval_end if x < m else interval_start) - x
                d = __c * e
            u = x + (d if abs(d) >= tol_radius else (tol_radius if d > 0 else -tol_radius))
            fu = function(u)
            # print(f'x: {u} result :{fu}')
            if fu <= fx:
                if u < x:
                    interval_end = x
                else:
                    interval_start = x
                v = w
                fv = fw
                w = x
                fw = fx
                x = u
                fx = fu
            else:
                if u < x:
                    interval_start = u
                else:
                    interval_end = u
                if fu <= fw or w == x:
                    v = w
                    fv = fw
                    w = u
                    fw = fu
                elif fu <= fv or v == x or v == w:
                    v = u
                    fv = fu
        else:
            return fx
    # raise Exception("max iterations reached")
    return fx


def approximate_local_min_axenie(interval_start: float,
                                 interval_end: float,
                                 function: Callable[[float], float],
                                 tolerance: float) -> float:
    """
    This function approximates the local minima of the given function in an interval. For more details look up ISBN 978-1-306-35261-1.

    :param tolerance: The tolerance for the approximation.
    :param interval_start: The included beginning of the interval.
    :param interval_end: the included end of the interval.
    :param function: A function defined on the interval. The local minima will be approximated for this function.
    """
    assert tolerance >= 0.0


    # print(f'interval_start: {interval_start} interval_end: {interval_end} tolerance: {tolerance}')
    current_start = interval_start
    current_end = interval_end
    c = interval_end

    f_start = function(current_start)
    f_end = function(current_end)

    f_c = f_end

    for i in range(0, 3000):
        if (f_end > 0.0 and f_c > 0.0) or (f_end < 0.0 and f_c < 0.0):
            c = current_start
            f_c = f_start
            e = d = current_end - current_start
        if abs(f_c) < abs(f_end):
            current_start = current_end
            current_end = c
            c = current_start
            f_start = f_end
            f_end = f_c
            f_c = f_start
        tol = 2.0 * __eps * abs(current_end) + 0.5 * tolerance
        xm = 0.5 * (c - current_end)
        if abs(xm) <= tol or f_end == 0.0:
            return current_end
        if abs(e) >= tol and abs(f_start) > abs(f_end):
            s = f_end / f_start
            if current_start == c:
                p = 2.0 * xm * s
                q = 1.0 - s
            else:
                q = f_start / f_c
                r = f_end / f_c
                p = s * (2.0 * xm * q * (q - r) - (current_end - current_start) * (r - 1.0))
                q = (q - 1.0) * (r - 1.0) * (s - 1.0)
            if p > 0.0:
                q = -1 * q
            p = abs(p)
            min_1 = 3.0 * xm * q - abs(tol * q)
            min_2 = abs(e * q)
            selected_min = min_1 if min_1 < min_2 else min_2
            if 2.0 * p < selected_min:
                e = d
                d = p / q
            else:
                d = xm
                e = d
        else:
            d = xm
            e = d
        current_start = current_end
        f_start = f_end
        if abs(d) > tol:
            current_end = current_end + d
        else:
            current_end = current_end + (abs(tol) if xm >= 0 else abs(tol) * -1)
        f_end = function(current_end)
    raise Exception("This should not happen. Check your values and think about increasing the allowed iterations.")


from .learning_rate_function import LearningRateFunction

from .const_lrf import ConstLRF
from .exp_e_decay_lrf import ExpEDecayLRF
from .linear_lrf import *
from .inv_time_adaptation import InverseTimeAdaptation



import time
import typing
from typing import Optional, Callable, Literal, Any

import numpy as np
from numpy import ndarray, dtype, float64
from tqdm import tqdm

from . import approximate_local_min_axenie
from .spice_net_hcm import SpiceNetHcm
from .spice_net_som import SpiceNetSom


def _norm(vec: np.array) -> np.array:
    return vec / vec[vec.argmax()]  # TODO: Doc
    # return vec / np.sum(vec)


def _partition_training_data(values: typing.Mapping[str, list[float | np.ndarray[any, np.dtype[np.float64]]]],
                             batch_size: int) -> tuple[
    dict[str, list[list[float | ndarray[Any, dtype[float64]]]]], int]:
    result = {}
    portion_number = 0
    for key, value_set in values.items():
        result[key] = [value_set[i:i + batch_size] for i in range(0, len(value_set), batch_size)]
        portion_number = len(result[key])
    return result, portion_number


class SpiceNet:
    def __init__(self,
                 correlation_matrix: SpiceNetHcm):
        self.__soms = correlation_matrix.get_soms()
        self.__correlation_matrix = correlation_matrix

    def get_correlation_matrix(self):
        return self.__correlation_matrix

    def decode(self,
               values: typing.Mapping[str, float | np.ndarray[any, np.dtype[np.float64]]],
               decoder: Literal["naive", "bert-optimizer"] = "bert-optimizer") -> float:
        existing_keys = values.keys()
        missing_key = None
        som: SpiceNetSom = None
        for som in self.__soms:
            if som.get_key() not in existing_keys:
                if missing_key is None:
                    missing_key = som.get_key()
                    som = som
                else:
                    raise ValueError(f"Found 2 missing keys: {missing_key} and {som.get_key()}")
        if som is None:
            raise ValueError(f"Value for all soms given, no need to decode")

        activation_vectors = {}
        for given_som in self.__soms:
            if given_som.get_key() == missing_key:
                continue
            activation_vectors[given_som.get_key()] = given_som.get_activation_vector(
                values[given_som.get_key()])
        match decoder:

            case "naive":
                should_activation: np.ndarray = self.__correlation_matrix.calculate_should_pattern(missing_key,
                                                                                                   activation_vectors)
                winner_index = should_activation.argmax()
                return som.naive_decode(should_activation[winner_index], winner_index)

            case "bert-optimizer":
                # input is given
                # cost = (normed(som_known(input) * correlation) - normed(som_searched(x)))²
                #      = (normed(should) - normed(som_searched(x)))²

                # cost = som_known(input) - (som_searched(x) * correlation)
                # better:
                # cost = som_known_winning(input) - (som_searched_winning(x) * correlation)
                # tol=(1.0e-6)*(limL + limH)/2.0;
                should_activations = self.__correlation_matrix.calculate_should_pattern(missing_key,
                                                                                        activation_vectors)
                # norm
                normed_should_activations = _norm(should_activations)
                winning_index = normed_should_activations.argmax()

                fn = lambda x: np.linalg.norm(
                    normed_should_activations - _norm(som.get_activation_vector(x))) ** 2  # TODO: Doc quadratic

                start_index = winning_index - 1 if winning_index > 0 else winning_index
                end_index = winning_index + 1 if winning_index < len(normed_should_activations) - 2 else winning_index
                start: float = som.get_as_matrix()[start_index][0]
                end: float = som.get_as_matrix()[end_index][0]
                #print(f'start_index: {start_index}, end_index: {end_index}, winning_index: {winning_index}, start: {start}, end: {end}')
                return approximate_local_min_axenie(start, end, fn, tolerance=abs((1.0e-6) * (start + end) / 2.0))

    def fit(self,
            values: typing.Mapping[str, list[float | np.ndarray[any, np.dtype[np.float64]]]],
            epochs_on_batch: int,
            batch_size: Optional[int] = None,
            after_batch_callback: Optional[Callable] = None,
            print_output: Optional[bool] = False):
        """
        Use this class to train and use a SpiceNet
        :param values:
        :param epochs_on_batch: Define how often a batch is used for training.
        :param batch_size: The values will be split in batches of this size for the training.
        (First the soms will train on ALL the batches data, then the correlation matrix will be trained.)
        :param after_batch_callback: This method will be called after each training with a batch. (Use it for plotting or what ever)
        :param print_output: This will toggle a progressbar implemented with tqdm and stops the computation times.
        :return:
        """
        som_elapsed_time = 0
        cm_elapsed_time = 0
        tmp: float = 0.0

        data_len = -1
        for som in self.__soms:
            if values[som.get_key()] is None:
                raise ValueError('Missing value for ' + som.get_key())
            if data_len != -1 and data_len != len(values[som.get_key()]):
                raise ValueError('The length of the data columns is not allways the same.')
            if data_len == -1:
                data_len = len(values[som.get_key()])

        keys = [som.get_key() for som in self.__soms]

        b_size = data_len if batch_size is None else batch_size
        partitions, partition_count = _partition_training_data(values, b_size)

        iterator = range(partition_count) if print_output is False else tqdm(range(partition_count), colour='green')
        for i in iterator:
            if print_output:
                tmp = time.time()
            for som in self.__soms:
                som.fit(partitions[som.get_key()][i], epochs_on_batch)

            if print_output:
                som_elapsed_time += time.time() - tmp

            if print_output:
                tmp = time.time()

            temp = {key: partitions[key][i] for key in keys}
            self.__correlation_matrix.fit(values=temp,
                                          epochs=epochs_on_batch)
            if print_output:
                cm_elapsed_time += time.time() - tmp

            if after_batch_callback is not None:
                after_batch_callback()

        if print_output:
            print(
                f'Time spend on the Components: \nSom: {som_elapsed_time} s | Convolution Matrix: {cm_elapsed_time} s')
